﻿/**
***  キャラクタアニメーションのための人体モデルの表現・基本処理 ライブラリ・サンプルプログラム
***  Copyright (c) 2015-, Masaki OSHITA (www.oshita-lab.org)
***  Released under the MIT license http://opensource.org/licenses/mit-license.php
**/

/**
***  動作接続・遷移アプリケーション
**/


// ライブラリ・クラス定義の読み込み
#include "SimpleHuman.h"
#include "MotionTransitionApp.h"
#include "BVH.h"
#include "Timeline.h"

// 標準算術関数・定数の定義
#define  _USE_MATH_DEFINES
#include <math.h>



//
//  動作のメタ情報を初期化
//
void  InitMotionInfo( MotionInfo * info, Motion * m )
{
	info->motion = m;
	info->begin_time = 0.0f;
	info->end_time = info->motion ? info->motion->GetDuration() : 0.0f;
	info->blend_end_time = info->begin_time;
	info->blend_begin_time = info->end_time;
}


//
//  サンプル動作セットの読み込み
//
const Skeleton *  LoadSampleMotions( vector< MotionInfo * > &  motion_list, const Skeleton * body )
{
	// アプリケーションのテストに使用する動作データの定義（BVHファイル名・キー時刻・描画色）
	// 各歩行動作の、右足が地面から離れる、右足が着く、左足が離れる、左足が着く、右足が離れるの５つのキー時刻を設定
	const int  num_motions = 3;
	const int  num_keytimes = 5; 
	const char *  sample_motions[ num_motions ] = {
		"sample_walking1.bvh", 
		"sample_walking2.bvh", 
		"sample_walking3.bvh" };
	const float  sample_keytimes[ num_motions ][ num_keytimes ] = {
		{ 2.35f, 3.00f, 3.08f, 3.68f, 3.74f },
		{ 1.30f, 2.07f, 2.12f, 2.88f, 2.94f },
		{ 1.20f, 2.00f, 2.08f, 2.80f, 2.86f }
	};
	const Color3f  sample_colors[] = {
		Color3f( 0.5f, 1.0f, 0.5f ), 
		Color3f( 0.5f, 0.5f, 1.0f ), 
		Color3f( 1.0f, 0.5f, 0.5f ) 
	};
	const char *  base_segments[ num_motions ] = {
		"LeftAnkle",
		"LeftAnkle",
		"LeftAnkle"
	};

	MotionInfo *  info = NULL;

	// 動作データの読み込み、動作情報の設定
	for ( int i=0; i<num_motions; i++ )
	{
		// BVH動作ファイルの読み込み
		BVH *  new_bvh = new BVH( sample_motions[ i ] );
		if ( !new_bvh || !new_bvh->IsLoadSuccess() )
			continue;

		// 骨格モデル・動作情報の生成（骨格モデルの生成は最初の一度のみ）
		Motion *  new_motion = CoustructBVHMotion( new_bvh, body );
		if ( !new_motion )
			continue;
		if ( !body )
			body = new_motion->body;

		// 動作のメタ情報の設定
		info = new MotionInfo();
		InitMotionInfo( info, new_motion );
		for ( int j = 0; j < num_keytimes; j++ )
			info->keytimes.push_back( sample_keytimes[ i ][ j ] );
		info->begin_time = info->keytimes[ 0 ];
		info->blend_end_time = info->keytimes[ 1 ];
		info->blend_begin_time = info->keytimes[ info->keytimes.size() - 2 ];
		info->end_time = info->keytimes[ info->keytimes.size() - 1 ];
		info->color = sample_colors[ i ];
		info->base_segment_no = FindSegment( new_motion->body, base_segments[ i ] );

		// 動作リストに追加
		motion_list.push_back( info );
	}

	return  body;
}



//
//  コンストラクタ
//
MotionTransitionApp::MotionTransitionApp()
{
	app_name = "Motion Transition";

	body = NULL;
	curr_posture = NULL;
	curr_motion = NULL;
	on_animation = true;
	animation_time = 0.0f;
	animation_speed = 1.0f;
	frame_no = 0;
	timeline = NULL;

	enable_connection = true;
	enable_transition = true;
}


//
//  デストラクタ
//
MotionTransitionApp::~MotionTransitionApp()
{
	for ( int i=0; motion_list.size(); i++ )
	{
		delete  motion_list[ i ]->motion;
		delete  motion_list[ i ];
	}
	motion_list.clear();

	if ( curr_posture )
		delete  curr_posture;
	if ( body )
		delete  body;
	if ( timeline )
		delete  timeline;
}


//
//  初期化
//
void  MotionTransitionApp::Initialize()
{
	GLUTBaseApp::Initialize();

	// サンプル動作セットの読み込み
	if ( motion_list.size() == 0 )
		body = LoadSampleMotions( motion_list );

	// 姿勢の初期化
	if ( body )
	{
		if ( curr_posture )
			delete  curr_posture;

		curr_posture = new Posture( body );
		InitPosture( *curr_posture, body );
	}

	// タイムライン描画機能の初期化
	timeline = new Timeline();
}


//
//  開始・リセット
//
void  MotionTransitionApp::Start()
{
	GLUTBaseApp::Start();
	
	// 動作データが正しく初期化されていなければ終了
	if ( motion_list.size() == 0 )
	{
		curr_motion = NULL;
		curr_motion_no = -1;
		return;
	}

	curr_motion_no = 0;
	next_motion_no = -1;
	waiting_motion_no = 0;
	waiting_motion_changed = true;

	const MotionInfo *  curr_motion_info = NULL;
	if ( motion_list.size() > 0 )
		curr_motion_info = motion_list[ curr_motion_no ];
	if ( curr_motion_info )
		curr_motion = curr_motion_info->motion;
	else
	{
		curr_motion = NULL;
		curr_motion_no = -1;
	}
	animation_time = 0.0f;
	frame_no = 0;
	figure_color = curr_motion_info->color;
	transition_count = 0;

	Point3f   init_pos( 0.0f, 0.0f, 0.0f );
	Matrix3f  init_ori;
	init_ori.rotY( 180.0f * M_PI / 180.0f );
	curr_motion_mat.set( init_ori, init_pos, 1.0f );

	curr_start_time = animation_time;
	transition_info.prev_motion = curr_motion_info;
	transition_info.next_motion = NULL;
	transition_info.next_begin_time = curr_motion_info->end_time - curr_motion_info->begin_time;
	transition_info.blend_begin_time = transition_info.next_begin_time;
	transition_info.blend_end_time = transition_info.next_begin_time;
	transition_info.prev_motion_mat = curr_motion_mat;
	transition_info.next_motion_mat = curr_motion_mat;

	Animation( 0.0f );
}


//
//  画面描画
//
void  MotionTransitionApp::Display()
{
	GLUTBaseApp::Display();

	// キャラクタを描画
	if ( curr_posture )
	{
		glColor3f( figure_color.x, figure_color.y, figure_color.z );
		DrawPosture( *curr_posture );
		DrawPostureShadow( *curr_posture, shadow_dir, shadow_color );
	}

	// タイムラインを描画
	if ( timeline )
		timeline->DrawTimeline();

	// 現在のモード、現在・次の再生動作、時間・フレーム番号を表示
	char  message[ 64 ];
	DrawTextInformation( 0, "Motion Transition" );
	if ( curr_motion_no != -1 )
	{
		if ( ( next_motion_no != -1 ) && ( next_motion_no != curr_motion_no ) )
			sprintf( message, "%s -> %s", motion_list[ curr_motion_no ]->motion->name.c_str(), motion_list[ next_motion_no ]->motion->name.c_str() );
		else
			sprintf( message, "%s", motion_list[ curr_motion_no ]->motion->name.c_str() );
		DrawTextInformation( 1, message );
	}
	if ( curr_motion )
		sprintf( message, "%.2f (%d)", animation_time, frame_no );
	else
		sprintf( message, "Failed to open the sample motions" );
	DrawTextInformation( 2, message );

	// 動作接続・遷移の設定を表示
	if ( !enable_connection || !enable_transition )
	{
		if ( !enable_connection && !enable_transition )
			sprintf( message, "Connection: Off, Transition: Off" );
		else if ( !enable_connection )
			sprintf( message, "Connection: Off" );
		else if ( !enable_transition )
			sprintf( message, "Transition: Off" );
		DrawTextInformation( 3, message );
	}
}


//
//  ウィンドウサイズ変更
//
void  MotionTransitionApp::Reshape( int w, int h )
{
	GLUTBaseApp::Reshape( w, h );

	// タイムラインの描画領域の設定（画面の下部に配置）
	if ( timeline )
		timeline->SetViewAreaBottom( 0, 0, 0, 3, 32, 2 );
}


//
//  マウスクリック
//
void  MotionTransitionApp::MouseClick( int button, int state, int mx, int my )
{
	GLUTBaseApp::MouseClick( button, state, mx, my );

	// 左ボタンが押されたら、次の再生動作を変更
	if ( ( button == GLUT_LEFT_BUTTON ) && ( state == GLUT_DOWN ) )
	{
		int  motion_no = ( curr_motion_no + 1 ) % 3;
		SetNextMotion( motion_no );
	}
}


//
//  キーボードのキー押下
//
void  MotionTransitionApp::Keyboard( unsigned char key, int mx, int my )
{
	GLUTBaseApp::Keyboard( key, mx, my );

	// 数字キーで次に実行する動作を指定
	if ( ( key >= '1' ) && ( key <= '9' ) )
	{
		int  no = key - '1';
		SetNextMotion( no );
	}

	// s キーでアニメーションの停止・再開
	if ( key == 's' )
		on_animation = !on_animation;

	// w キーでアニメーションの再生速度を変更
	if ( key == 'w' )
		animation_speed = ( animation_speed == 1.0f ) ? 0.1f : 1.0f;

	// n キーで次のフレーム
	if ( ( key == 'n' ) && !on_animation && curr_motion )
	{
		on_animation = true;
		Animation( curr_motion->interval );
		on_animation = false;
	}

	// d キーで動作接続・遷移の設定を変更
	if ( key == 'd' )
	{
		if ( enable_connection && enable_transition )
			enable_transition = false;
		else if ( enable_connection && !enable_transition )
			enable_connection = false;
		else if ( !enable_connection && !enable_transition )
		{
			enable_connection = true;
			enable_transition = true;
		}
	}
}


//
//  キーボードの特殊キー押下
//
void  MotionTransitionApp::KeyboardSpecial( unsigned char key, int mx, int my )
{
	// カーソル上キーが押されたら、次の再生動作を変更
	if ( key == GLUT_KEY_UP )
	{
		SetNextMotion();
	}
}


//
//  アニメーション処理
//
void  MotionTransitionApp::Animation( float delta )
{
	// アニメーション再生中でなければ終了
	if ( !on_animation )
		return;

	//  動作再生処理（動作接続のみを考慮、動作遷移は考慮しない）
	if ( !enable_transition )
		AnimationWithConnection( delta );
	//  動作再生処理（動作接続・遷移を考慮）
	else
		AnimationWithConnectionTransition( delta );

	// 注視点を更新
	view_center.set( curr_posture->root_pos.x, 0.0f, curr_posture->root_pos.z );
}


//
//  次の動作を変更
//
void  MotionTransitionApp::SetNextMotion( int no )
{
	if ( motion_list.size() == 0 )
		return;

	// 次の動作が指定されなかった場合は、現在の再生動作の次の番号の動作を次の動作とする
	if ( no < 0 )
		no = ( curr_motion_no + 1 ) % motion_list.size();
	else
		no = no % motion_list.size();

	// 再生待ちの動作に設定
	waiting_motion_no = no;

	// 再生待ちの動作が変更されたかどうかのフラグを設定
	waiting_motion_changed = true;
}


//
//  動作再生処理（動作接続を考慮、動作遷移は考慮しない）
//
void  MotionTransitionApp::AnimationWithConnection( float delta )
{
	// 現在・次の再生動作、待ち動作の情報
	MotionInfo *  curr_motion_info = NULL;
	MotionInfo *  waiting_motion_info = NULL;
	MotionInfo *  next_motion_info = NULL;
	Motion *  next_motion = NULL;

	// 現在の動作の開始時刻（curr_start_time）を基準とする、現在の動作の経過時間
	float  local_time = 0.0f;

	// 動作接続の初期化が完了したかどうかの状態
	bool  connection_initiated = false;

	// 動作接続が完了したかどうかの状態
	bool  connection_completed = false;


	// アニメーションの時間を進める
	animation_time += delta * animation_speed;

	// 現在の再生動作が設定されていなければ終了
	if ( !curr_motion )
	{
		frame_no = 0;
		return;
	}

	// 現在の再生動作の情報を取得
	curr_motion_info = motion_list[ curr_motion_no ];

	// 現在の動作の開始時刻（curr_start_time）からの経過時間を計算
	local_time = animation_time - curr_start_time;

	// 現在の動作のフレーム番号を計算（表示用）
	frame_no = local_time / curr_motion->interval;


	// 動作接続・遷移の初期化
	if ( waiting_motion_changed )
	{
		// 実行待ちの動作を取得
		waiting_motion_info = motion_list[ waiting_motion_no ];

		// 動作接続の初期化
		connection_initiated = InitConnection( curr_motion_info, waiting_motion_info, curr_motion_mat, local_time, &transition_info );

		// 動作接続・遷移の初期化に成功したら、次の動作を更新
		if ( connection_initiated )
		{
			// 実行待ちの動作を次の動作として設定
			next_motion_no = waiting_motion_no;

			// 実行待ちの動作を初期化
			waiting_motion_no = -1;
			waiting_motion_changed = false;

			// タイムラインへの前後の動作・遷移区間の設定
			InitTimeline( *timeline, curr_start_time, transition_info );
		}
	}

	// 次の再生動作の情報を取得
	next_motion_info = motion_list[ next_motion_no ];
	next_motion = next_motion_info->motion;

	// 動作接続・遷移の姿勢計算
	connection_completed = GetPostureInConnection( &transition_info, local_time, curr_posture );

	// 姿勢の描画色を設定
	if ( !connection_completed )
		figure_color = curr_motion_info->color;
	else
		figure_color = next_motion_info->color;

	// タイムラインへの現在時刻・時間範囲の設定
	UpdateTimeline( *timeline, animation_time );


	// 動作接続の完了処理
	if ( connection_completed )
	{
		// 現在の動作を次の動作に切り替え
		curr_motion = next_motion;
		curr_motion_no = next_motion_no;

		// 現在の動作の再生開始時刻（グローバル時刻）を更新
		curr_start_time = curr_start_time + transition_info.next_begin_time;

		// 再生動作の位置・向きを合わせるための変換行列を更新
		curr_motion_mat = transition_info.next_motion_mat;

		// 次の動作を初期化
		next_motion_no = -1;

		// 実行待ち動作が未設定であれば、現在の動作を繰り返すように設定
		if ( waiting_motion_no == -1 )
			waiting_motion_no = curr_motion_no;

		// 待ち動作のフラグをクリア
		waiting_motion_changed = true;

		// 動作遷移回数のカウントを加算（タイムライン表示用）
		transition_count ++;
	}
}


//
//  動作再生処理（動作接続・遷移を考慮）
//
void  MotionTransitionApp::AnimationWithConnectionTransition( float delta )
{
	// 現在・次の再生動作、待ち動作の情報
	MotionInfo *  curr_motion_info = NULL;
	MotionInfo *  waiting_motion_info = NULL;
	MotionInfo *  next_motion_info = NULL;
	Motion *  next_motion = NULL;

	// 現在の動作の開始時刻（curr_start_time）を基準とする、現在の動作の経過時間
	float  local_time = 0.0f;

	// 動作接続・遷移の初期化が完了したかどうかの状態
	bool  transition_initiated = false;

	// 動作接続・遷移のブレンド区間が完了したかどうかの状態
	bool  transition_completed = false;

	// 動作遷移中の前後の動作の姿勢のブレンド比率
	float  blend_ratio = 0.0f;


	// アニメーションの時間を進める
	animation_time += delta * animation_speed;

	// 現在の再生動作が設定されていなければ終了
	if ( !curr_motion )
	{
		frame_no = 0;
		return;
	}

	// 現在の再生動作の情報を取得
	curr_motion_info = motion_list[ curr_motion_no ];

	// 現在の動作の開始時刻（curr_start_time）からの経過時間を計算
	local_time = animation_time - curr_start_time;

	// 現在の動作のフレーム番号を計算（表示用）
	frame_no = local_time / curr_motion->interval;


	// 動作接続・遷移の初期化
	if ( waiting_motion_changed )
	{
		// 実行待ちの動作を取得
		waiting_motion_info = motion_list[ waiting_motion_no ];

		// 動作接続・遷移の初期化
		transition_initiated = InitTransition( curr_motion_info, waiting_motion_info, curr_motion_mat, local_time, &transition_info );

		// 動作接続・遷移の初期化に成功したら、次の動作を更新
		if ( transition_initiated )
		{
			// 実行待ちの動作を次の動作として設定
			next_motion_no = waiting_motion_no;

			// 実行待ちの動作を初期化
			waiting_motion_no = -1;
			waiting_motion_changed = false;

			// タイムラインへの前後の動作・遷移区間の設定
			InitTimeline( *timeline, curr_start_time, transition_info );
		}
	}

	// 次の再生動作の情報を取得
	if ( next_motion_no != -1 )
	{
		next_motion_info = motion_list[ next_motion_no ];
		next_motion = next_motion_info->motion;
	}

	// 動作接続・遷移の姿勢計算
	transition_completed = GetPostureInTransition( &transition_info, local_time, curr_posture, &blend_ratio );

	// 姿勢の描画色を設定
	if ( blend_ratio > 0.0f )
		figure_color.scaleAdd( blend_ratio, next_motion_info->color - curr_motion_info->color, curr_motion_info->color );
	else
		figure_color = curr_motion_info->color;

	// タイムラインへの現在時刻・時間範囲の設定
	UpdateTimeline( *timeline, animation_time );


	// 動作接続・遷移の動作ブレンドの完了処理
	if ( transition_completed )
	{
		// 現在の動作を次の動作に切り替え
		curr_motion = next_motion;
		curr_motion_no = next_motion_no;

		// 現在の動作の再生開始時刻（グローバル時刻）を更新
		curr_start_time = curr_start_time + transition_info.next_begin_time;

		// 再生動作の位置・向きを合わせるための変換行列を更新
		curr_motion_mat = transition_info.next_motion_mat;

		// 次の動作を初期化
		next_motion_no = -1;

		// 実行待ち動作が未設定であれば、現在の動作を繰り返すように設定
		if ( waiting_motion_no == -1 )
			waiting_motion_no = curr_motion_no;

		// 待ち動作のフラグをクリア
		waiting_motion_changed = true;

		// 動作遷移回数のカウントを加算（タイムライン表示用）
		transition_count ++;
	}
}


//
//  動作接続の初期化
//
bool  MotionTransitionApp::InitConnection( 
	const MotionInfo * prev_motion, const MotionInfo * next_motion, const Matrix4f & prev_motion_mat, float curr_local_time, 
	 MotionTransitionInfo * transition )
{
	// 現在の動作の開始時刻（curr_start_time）を基準とする、次の動作の開始時刻
	float  next_begin_time = 0.0f;

	// 次の動作の位置・向きを前の動作の位置・向きに合わせるための変換行列
	Matrix4f  next_motion_mat;


	// 次の動作を開始する時刻（現在の動作の開始時刻 curr_start_time を基準とするローカル時刻）
	next_begin_time = prev_motion->end_time - prev_motion->begin_time;

	// 現在の動作の終了姿勢（ワールド座標系）を取得
	Posture   prev_end_posture( prev_motion->motion->body );
	prev_motion->motion->GetPosture( prev_motion->end_time, prev_end_posture );
	TransformPosture( prev_motion_mat, prev_end_posture );

	// 次の動作の開始姿勢（動作のワールド座標系）を取得
	Posture   next_begin_posture( next_motion->motion->body );
	next_motion->motion->GetPosture( next_motion->begin_time, next_begin_posture );

	// 現在の動作の終了姿勢と次の動作の開始姿勢の姿勢の位置・向きを合わせるための変換行列を計算
	//（curr_posture の位置・向きを、next_motion_posture の位置向きに合わせるための変換行列 trans_mat を計算）
	ComputeConnectionTransformation( prev_end_posture, 0.0f, next_begin_posture, 180.0f, 
		prev_motion->base_segment_no, next_motion_mat );

	// 動作接続・遷移情報を設定
	transition->prev_motion = prev_motion;
	transition->next_motion = next_motion;
	transition->next_begin_time = next_begin_time;
	transition->blend_begin_time = next_begin_time;
	transition->blend_end_time = next_begin_time;
	transition->prev_motion_mat = prev_motion_mat;
	transition->next_motion_mat = next_motion_mat;

	return  true;
}


//
//  動作接続の姿勢計算
//
bool  MotionTransitionApp::GetPostureInConnection( 
	const MotionTransitionInfo * transition, float curr_local_time, Posture * posture )
{
	// 次の動作の現在時刻（次の動作の開始時刻 next_motion_begin_time を基準とするローカル時刻）
	float  next_motion_local_time = 0.0f;

	// 動作接続後かどうかを判定
	bool  on_motion_connection = false;
	if ( curr_local_time > transition->next_begin_time )
		on_motion_connection = true;

	// 動作接続前は、前の動作から姿勢を取得
	if ( !on_motion_connection )
	{
		// 現在の動作から現在時刻の姿勢を取得
		transition->prev_motion->motion->GetPosture( 
			curr_local_time + transition->prev_motion->begin_time, *posture );

		// 現在の動作の姿勢の位置・向きに変換行列を適用
		TransformPosture( transition->prev_motion_mat, *posture );
	}
	// 動作接続後は、次の動作から姿勢を取得
	else
	{
		// 次の動作の現在時刻を計算（次の動作の開始時刻 next_motion_begin_time を基準とするローカル時刻）
		next_motion_local_time = curr_local_time - transition->next_begin_time;

		// 次の動作から現在時刻の姿勢を取得
		transition->next_motion->motion->GetPosture( 
			next_motion_local_time + transition->next_motion->begin_time, *posture );

		// 次の動作の姿勢の位置・向きに変換行列を適用
		TransformPosture( transition->next_motion_mat, *posture );
	}

	// 動作接続が完了したら、戻り値として真を返す
	return  on_motion_connection;
}


//
//  動作接続・遷移の初期化
//
bool  MotionTransitionApp::InitTransition( 
	const MotionInfo * prev_motion, const MotionInfo * next_motion, const Matrix4f & prev_motion_mat, float curr_local_time, 
	MotionTransitionInfo * transition )
{
	// 現在の動作の開始時刻（curr_start_time）を基準とする、次の動作の開始時刻
	float  next_begin_time = 0.0f;

	// 現在の動作の開始時刻（curr_start_time）を基準とする、動作ブレンドの開始・終了時刻
	float  blend_begin_time = 0.0f;
	float  blend_end_time = 0.0f;

	// 前後の動作の接続の基準体節
	int  base_segment_no = -1;

	// 次の動作の位置・向きを前の動作の位置・向きに合わせるための変換行列
	Matrix4f  next_motion_mat;
	next_motion_mat.set( prev_motion_mat );

	// ※ レポート課題

	// 次の動作を開始する時刻（現在の動作の開始時刻 curr_start_time を基準とするローカル時刻）
//	next_begin_time = ???;

	// 動作遷移のための動作ブレンドを行う開始時刻（現在の動作の開始時刻 curr_start_time を基準とするローカル時刻）
//	blend_begin_time = ???;

	// 動作遷移のための動作ブレンドを行う終了時刻（現在の動作の開始時刻 curr_start_time を基準とするローカル時刻）
//	blend_end_time = ???;

	// 現在時間が遷移開始時間を過ぎている場合は、動作接続・遷移の開始は行わない
	if ( curr_local_time >= next_begin_time )
		return  false;

	// 現在の動作の終了姿勢（ワールド座標系）を取得
	Posture   prev_end_posture( prev_motion->motion->body );
	float  prev_local_time =  blend_begin_time + prev_motion->begin_time;
	prev_motion->motion->GetPosture( prev_local_time, prev_end_posture );
	TransformPosture( prev_motion_mat, prev_end_posture );

	// 次の動作の開始姿勢（動作のワールド座標系）を取得
	Posture   next_begin_posture( next_motion->motion->body );
	float  next_local_time =  blend_end_time - next_begin_time + next_motion->begin_time;
	next_motion->motion->GetPosture( next_local_time, next_begin_posture );

	// 接続の基準とする体節が未設定であれば、前の動作の情報を使用
	if ( base_segment_no == -1 )
		base_segment_no = prev_motion->base_segment_no;

	// 現在の動作の終了姿勢と次の動作の開始姿勢の姿勢の位置・向きを合わせるための変換行列を計算
	//（curr_posture の位置・向きを、next_motion_posture の位置向きに合わせるための変換行列 trans_mat を計算）
	ComputeConnectionTransformation( prev_end_posture, 0.0f, next_begin_posture, 180.0f, 
		base_segment_no, next_motion_mat );

	// 動作接続・遷移情報を設定
	transition->prev_motion = prev_motion;
	transition->next_motion = next_motion;
	transition->next_begin_time = next_begin_time;
	transition->blend_begin_time = blend_begin_time;
	transition->blend_end_time = blend_end_time;
	transition->prev_motion_mat = prev_motion_mat;
	transition->next_motion_mat = next_motion_mat;

	return  true;
}


//
//  動作接続・遷移の姿勢計算
//
bool  MotionTransitionApp::GetPostureInTransition( 
	const MotionTransitionInfo * transition, float curr_local_time, 
	Posture * posture, float * output_blend_ratio )
{
	// 次の動作の現在時刻（次の動作の開始時刻 next_motion_begin_time を基準とするローカル時刻）
	float  next_motion_local_time = 0.0f;

	// 姿勢補間の割合
	float  blend_ratio = 0.0f;

	// 前後の動作の姿勢
	Posture   prev_motion_posture;
	Posture   next_motion_posture;


	// 動作遷移開始後かどうかを判定
	bool  on_motion_transition = false;
	if ( curr_local_time > transition->blend_begin_time )
		on_motion_transition = true;

	// 現在の動作から現在時刻の姿勢を取得
	transition->prev_motion->motion->GetPosture( 
		curr_local_time + transition->prev_motion->begin_time, prev_motion_posture );

	// 現在の動作の姿勢の位置・向きに変換行列を適用
	TransformPosture( transition->prev_motion_mat, prev_motion_posture );

	// 動作接続・遷移のためのブレンド中は、次の動作の姿勢とブレンド
	if ( on_motion_transition && transition->next_motion )
	{
		// 次の動作の現在時刻を計算（次の動作の開始時刻 next_motion_begin_time を基準とするローカル時刻）
		next_motion_local_time = curr_local_time - transition->next_begin_time;

		// 次の動作から現在時刻の姿勢を取得
		transition->next_motion->motion->GetPosture( 
			next_motion_local_time + transition->next_motion->begin_time, next_motion_posture );

		// 次の動作の姿勢の位置・向きに変換行列を適用
		TransformPosture( transition->next_motion_mat, next_motion_posture );

		// ※ レポート課題

		// ブレンド比率（補間の重み）を計算
//		blend_ratio = ???;

		// ブレンド比率（補間の重み）が範囲外の値になった場合への対応
		if ( blend_ratio > 1.0f )
			blend_ratio = 1.0f;
		if ( blend_ratio < 0.0f )
			blend_ratio = 0.0f;

		// 前後の動作の姿勢を補間
		MyPostureInterpolation( prev_motion_posture, next_motion_posture, blend_ratio, *posture );
	}
	else
	{
		// 前の動作の姿勢を出力姿勢として設定
		*posture = prev_motion_posture;
	}

	// 前後の動作の姿勢のブレンド比率を出力（表示用）
	if ( output_blend_ratio )
		*output_blend_ratio = blend_ratio;

	// 動作遷移が完了したら、戻り値として真を返す
	return  ( blend_ratio == 1.0f );
}


//
//  タイムラインへの現在動作・次の動作・遷移区間の設定
//
void  MotionTransitionApp::InitTimeline( Timeline & timeline, float prev_start_time, const MotionTransitionInfo & trans  )
{
	Color4f  prev_color, next_color, trans_color;

	const MotionInfo *  prev_motion = trans.prev_motion;
	const MotionInfo *  next_motion = trans.next_motion;

	// タイムラインに各動作を表す要素を追加
	while( timeline.GetNumElements() < 3 )
		timeline.AddElement( 0.0f, 1.0f, Color4f( 1.0f, 1.0f, 1.0f, 1.0f ), NULL, timeline.GetNumElements() );

	// 前の動作・次の動作を配置するトラック番号を決定
	int  prev_track_no = 0;
	int  next_track_no = 2;
	if ( transition_count % 2 == 1 )
	{
		prev_track_no = 2;
		next_track_no = 0;
	}

	// 前の動作の要素の情報を設定
	if ( trans.prev_motion )
	{
		timeline.SetElementEnable( 0, true );
		timeline.SetElementTime( 0, prev_start_time, 
			prev_start_time + ( prev_motion->end_time - prev_motion->begin_time ) );
		prev_color.set( prev_motion->color.x, prev_motion->color.y, prev_motion->color.z, 1.0f );
		timeline.SetElementColor( 0, prev_color );
		timeline.SetElementText( 0, prev_motion->motion->name.c_str() );
		timeline.SetElementTrackNo( 0, prev_track_no );
	}
	else
	{
		timeline.SetElementEnable( 0, false );
	}

	// 遷移後の動作の要素の情報を設定
	if ( trans.next_motion )
	{
		timeline.SetElementEnable( 2, true );
		timeline.SetElementTime( 2, prev_start_time + trans.next_begin_time, 
			prev_start_time + trans.next_begin_time + ( next_motion->end_time - next_motion->begin_time ) );
		next_color.set( next_motion->color.x, next_motion->color.y, next_motion->color.z, 1.0f );
		timeline.SetElementColor( 2, next_color );
		timeline.SetElementText( 2, next_motion->motion->name.c_str() );
		timeline.SetElementTrackNo( 2, next_track_no );
	}
	else
	{
		timeline.SetElementEnable( 2, false );
	}

	// 遷移区間の要素の情報を設定
	if ( trans.next_motion )
	{
		timeline.SetElementEnable( 1, true );
		timeline.SetElementTime( 1, prev_start_time + trans.blend_begin_time, prev_start_time + trans.blend_end_time );
		timeline.SetElementColor( 1, prev_color, next_color );
		timeline.SetElementText( 1, "transition" );
		timeline.SetElementTrackNo( 1, 1 );
	}
	else
	{
		timeline.SetElementEnable( 1, false );
	}

	// 各動作のブレンド区間の表示に必要な区間を初期化
	int  size = 4;
	while( timeline.GetNumSubElements() < size )
		timeline.AddSubElement( 0, 0.0f, 1.0f, 0.0f, 1.0f, Color4f( 1.0f, 1.0f, 1.0f, 1.0f ) );

	// 各動作の開始・終了ブレンド区間の要素の情報を設定
	float  base_time;
	int  count = 0;
	base_time = prev_start_time - prev_motion->begin_time;
	prev_color.set( prev_motion->color.x + 0.1f, prev_motion->color.y + 0.1f, prev_motion->color.z + 0.1f, 1.0f );
	timeline.SetSubElementEnable( count, true );
	timeline.SetSubElementParent( count, 0 );
	timeline.SetSubElementTime( count, base_time + prev_motion->begin_time, base_time + prev_motion->blend_end_time );
	timeline.SetSubElementColor( count, prev_color );
	count ++;
	timeline.SetSubElementEnable( count, true );
	timeline.SetSubElementParent( count, 0 );
	timeline.SetSubElementTime( count, base_time + prev_motion->blend_begin_time, base_time + prev_motion->end_time );
	timeline.SetSubElementColor( count, prev_color );
	count ++;

	base_time = prev_start_time + trans.next_begin_time - next_motion->begin_time;
	next_color.set( next_motion->color.x + 0.1f, next_motion->color.y + 0.1f, next_motion->color.z + 0.1f, 1.0f );
	timeline.SetSubElementEnable( count, true );
	timeline.SetSubElementParent( count, 2 );
	timeline.SetSubElementTime( count, base_time + next_motion->begin_time, base_time + next_motion->blend_end_time );
	timeline.SetSubElementColor( count, next_color );
	count ++;
	timeline.SetSubElementEnable( count, true );
	timeline.SetSubElementParent( count, 2 );
	timeline.SetSubElementTime( count, base_time + next_motion->blend_begin_time, base_time + next_motion->end_time );
	timeline.SetSubElementColor( count, next_color );
	count ++;
}


//
//  タイムラインへの現在時刻の設定
//
void  MotionTransitionApp::UpdateTimeline( Timeline & timeline, float curr_time )
{
	// 現在時刻のラインの情報を設定
	while( timeline.GetNumLines() < 1 )
		timeline.AddLine( curr_time, Color4f( 1.0f, 0.0f, 0.0f, 1.0f ) );
	timeline.SetLineTime( 0, curr_time );

	// 描画時間範囲の情報を設定
	timeline.SetTimeRange( curr_time - 2.0f, curr_time + 3.0f );
}



//
//  以下、補助処理
//


//
//  変換行列の水平向き（方位角）成分を計算（Ｚ軸の正の方向を０とする時計回りの角度を -180～180 の範囲で求める）
//
float  ComputeOrientation( const Matrix3f & ori )
{
	Vector3f  fig_ori;
	ori.getColumn( 2, &fig_ori );
	return  atan2( fig_ori.x, fig_ori.z ) * 180.0f / M_PI;
}


//
//  ２つの姿勢の位置・向きを合わせるための変換行列を計算
// （next_frame の位置・向きを、prev_frame の位置・向きに合わせるための変換行列 trans_mat を計算）
//
void  ComputeConnectionTransformation( const Matrix4f & prev_frame, const Matrix4f & next_frame, Matrix4f & trans_mat )
{
	// ※ レポート課題

}


//
//  ２つの姿勢の位置・向きを合わせるための変換行列を計算
//  ２つの姿勢の腰の水平位置・水平向きを合わせる
// （next_pose の腰の水平位置・水平向きを、prev_pose の腰の水平位置・水平向きに合わせる）
//
void  ComputeConnectionTransformation( const Posture & prev_pose, const Posture & next_pose, Matrix4f & trans_mat )
{
	// 前の姿勢の腰の位置・向きを取得
	Matrix4f  curr_end_frame;
	curr_end_frame.set( prev_pose.root_ori, prev_pose.root_pos, 1.0f );

	// 次の姿勢の腰の位置・向きを取得
	Matrix4f  next_begin_frame;
	next_begin_frame.set( next_pose.root_ori, next_pose.root_pos, 1.0f );

	// 次の姿勢の腰の位置・向きを前の姿勢の腰の位置・向きに合わせるための変換行列を計算
	ComputeConnectionTransformation( curr_end_frame, next_begin_frame, trans_mat );
}


//
//  ２つの姿勢の位置・向きを合わせるための変換行列を計算
//  ２つの姿勢の任意の体節の水平位置と、入力として与えられた水平向きを合わせる
// （next_pose の基準部位 base_segment の水平位置を prev_pose の水平位置に合わせて、水平向き next_ori を 水平向き prev_ori に合わせる）
//
void  ComputeConnectionTransformation( const Posture & prev_pose, float prev_ori, 
	const Posture & next_pose, float next_ori, int base_segment, Matrix4f & trans_mat )
{
	Vector3f  prev_pos, next_pos;
	Matrix3f  prev_rot, next_rot;
	Matrix4f  prev_frame, next_frame;

	// 基準部位としてルート体節が指定された場合は、２つの姿勢のルートの位置を取得
	if ( base_segment == 0 )
	{
		// ※レポート課題
//		prev_pos = ???;
//		next_pos = ???;
	}
	// 基準部位としてルート体節以外が指定された場合は、順運動学計算により、２つの姿勢の指定部位の位置を計算
	else
	{
		// ※レポート課題
//		prev_pos = ???;
//		next_pos = ???;
	}

	// 前の姿勢の位置・向きを表す変換行列を計算
	// ※レポート課題
//	prev_rot = ???;
	prev_frame.set( prev_rot, prev_pos, 1.0f );

	// 次の姿勢の位置・向きを表す変換行列を計算
	// ※レポート課題
//	next_rot = ???;
	next_frame.set( next_rot, next_pos, 1.0f );

	// 座標変換を計算
	// ※レポート課題
//	trans_mat = ???;

	// 上記の処理が未実装であれば、動作データの水平向きや基準部位の情報を使用しない、標準的な処理を呼び出す
	ComputeConnectionTransformation( prev_pose, next_pose, trans_mat );
}


//
//  姿勢の位置・向きに変換行列を適用
//
void  TransformPosture( const Matrix4f & trans, Posture & posture )
{
	// 変換行列を適用
	Matrix3f  rot;
	trans.get( &rot );
	posture.root_ori.mul( rot, posture.root_ori );
	trans.transform( &posture.root_pos );
}


